#include "xparameters.h"
#include "stdio.h"
#include "xutil.h"

#define LAUNCH          1
#define RUN             0

#include "xuartlite_l.h"
#define UART_BASEADDR STDOUT_BASEADDRESS 

#define input()             ( (unsigned) inbyte () )

#define g_sum (((volatile unsigned char *) ( \
            (XPAR_DDR2_SDRAM_MPMC_HIGHADDR) & ~0xf))[0])


static void crash ( void )
{
    while(1) {}
}

static unsigned get_hex_nibble ( void )
{
    unsigned ch = input () ;
    ch -= 'a' ;
    if ( ch < 6 ) return ch + 10 ;
    ch += 'a' - 'A' ;
    if ( ch < 6 ) return ch + 10 ;
    ch += 'A' - '0' ;
    if ( ch < 10 ) return ch ;

    return 0;
}

static unsigned get_hex_byte ( void )
{
    unsigned rc ;
    rc = ( get_hex_nibble () << 4 ) | get_hex_nibble () ;
    g_sum += rc ;
    return rc ;
}

static void printhex(unsigned x)
{
    char c[9];
    unsigned i, y;

    for (i = 0; i < 8; i++) {
        y = x >> 28;
        if (y < 10) {
            c[i] = y + '0';
        } else {
            c[i] = y + 'A' - 10;
        }
        x = (x << 4);
    }
    c[8] = '\0';
    print(c);
}

void stage2(unsigned reloc);

int main(void)
{
    unsigned i;
    volatile unsigned * from = (unsigned *) 0;
    volatile unsigned * to = (unsigned *) 
            ((((XPAR_DDR2_SDRAM_MPMC_HIGHADDR) -
                (XPAR_DLMB_CNTLR_HIGHADDR)) - 0x1000) & ~0xff);
    unsigned target;
    unsigned reloc;

    XUartLite_mSetControlReg(UART_BASEADDR, (XUL_CR_ENABLE_INTR    |
                                           XUL_CR_FIFO_RX_RESET  |
                                           XUL_CR_FIFO_TX_RESET));

    for (i = 0; i < 5; i++) {
        volatile int x, y;
        y = 0;
        for (x = 100000; x > 0; x--) { y++; }

        //while (XUartLite_mIsTransmitFull(UART_BASEADDR));
        //XUartLite_SendByte(UART_BASEADDR, 'x');
        print("/");
    }
        
    // relocate myself to SDRAM
    print("\r\nrelocating...");

    for (i = 0; i < (XPAR_DLMB_CNTLR_HIGHADDR / 4); i++) {
        to[i] = from[i];
    }
    //print(".");

    for (i = 0; i < (XPAR_DLMB_CNTLR_HIGHADDR / 4); i++) {
        if (to[i] != from[i]) {
            print("verify failed.\r\n");
            crash();
            return 0;
        }
    }
    print(".");

    target = 0;
    asm volatile(
            ".global stage2\n"
            "addi %0, r0, stage2\n" : "=r"(target) );

    reloc = (unsigned) to - (unsigned) from;
    target += reloc;

    asm volatile(
            "addi r1, %1, -0x100\n"
            "rtsd %0, 0\n"
            "or r5, r0, %1\n" : : "r"(target), "r"(reloc) : "r5");
    crash();
    return 0;
}


#define RS(name) ((char *) (((unsigned) (name)) + reloc))
void stage2(unsigned reloc)
{
    unsigned        state, byte_count, address;
    unsigned        high_address, rec_type, entrance;
    volatile unsigned char * memory ;
    unsigned char   byte ;
    unsigned i;
    volatile unsigned * from = (unsigned *) 0;

    print(RS("ok"));
    for (i = 0; i < (XPAR_DLMB_CNTLR_HIGHADDR / 4); i++) {
        from[i] = ~0;
    }
    print(RS("\r\nsend hex:\r\n"));

    high_address = 0 ;
    entrance = 0 ;
    state = RUN ;
    while (state == RUN) {   
        while ( input () != ':' ) {}

        g_sum = 0 ;
        byte_count = get_hex_byte () ;
        address = get_hex_byte () << 8 ;
        address |= get_hex_byte () | high_address ;
        memory = (unsigned char *) address ;
        rec_type = get_hex_byte () ;
        if (rec_type == 0) { /* DO NOT USE switch */
            /* Data */
            print(RS("\r"));
            printhex((unsigned) memory);
            while (byte_count != 0) {
                byte = (unsigned char) get_hex_byte () ;
                memory[0] = byte;
                memory[1] = ~byte;
                memory[2] = 0xff;
                memory[3] = 0xab;
                if (memory[0] != byte) {
                    print(RS(" <-- Readback failed!\r\n"));
                    printhex(byte);
                    for (i = 0; i < 8; i++) {
                        printhex(memory[i]);
                    }
                    crash() ;
                }
                byte_count -- ;
                memory ++ ;
            }
        } else if (rec_type == 1) {
            /* EOF */
            state = LAUNCH ;
        } else if ((rec_type == 2) || (rec_type == 4)) {
            /* Either: (2) 16 bit segment address */
            /* or: (4) Address high word */
            high_address = 0 ;
            while ( byte_count != 0 )
            {
                high_address = ( high_address << 8 ) | get_hex_byte () ;
                byte_count -- ;
            }
            if ( rec_type == 2 )    /* 16 bit segment address */
            {
                high_address = high_address << 4 ;
            } else {                /* address high word */
                high_address = high_address << 16 ;
            }
        } else if (rec_type == 3) {    
            /* CS:IP value: entrance point */
            byte_count -= 4 ;
            entrance = (( get_hex_byte () << 8 ) | get_hex_byte () ) << 4 ;
            entrance += (( get_hex_byte () << 8 ) | get_hex_byte () ) ;
        } else if (rec_type == 5) {    
            /* Linear value: entrance point */
            entrance = 0 ;
            while (byte_count != 0) {
                entrance = ( entrance << 8 ) | get_hex_byte () ;
                byte_count -- ;
            }
        } else {
            print(RS("\r\nUnknown ihex record type."));
            crash();
        }
        while ( byte_count != 0 )
        {
            (void) get_hex_byte () ;
            byte_count -- ;
        }
        (void) get_hex_byte () ; /* checksum */
        if ((g_sum & 0xff) != 0) {
            printhex(memory);
            printhex(g_sum);
            print(RS("\r\nChecksum failed."));
            crash();
        }
    }
    print(RS("\r"));
    printhex(entrance);
    print(RS(" - boot\r\n"));
    asm volatile("rtsd %0, 0\n"
            "nop\n" : : "r"(entrance));
    crash();
}

