/* RTOS low-level code */
/* Cache variant */

#define FROM_ASM
#include "rtos.h"



rmsr_offset = 0x00
r1_offset = 0x04
r2_offset = 0x08
r3_offset = 0x0c
r4_offset = 0x10
r5_offset = 0x14
r6_offset = 0x18
r7_offset = 0x1c
r8_offset = 0x20
r9_offset = 0x24
r10_offset = 0x28
r11_offset = 0x2c
r12_offset = 0x30
r13_offset = 0x34
r14_offset = 0x38
r15_offset = 0x3c
r15_b_offset = 0x40
r18_offset = 0x44
r19_offset = 0x48
r1_b_offset = 0x4c
switch_area_size = r1_b_offset + 4

# defined here
.global reset
.global exit
.global _exit
.global _open
.global notify_task_name
.global kernel_idle
.global init_asm_start
.global cs_measure_ep
.global cs_measure_sample
.global measurement_patch_point

# defined elsewhere
.global call_function
.global scheduler
.global main
.global print
.global sw_log_event
.global cleardown
.global kernel_entry
.global task_func_is_zero
.global __end_spm

	.section	.kernelstuff.t,"ax",@progbits
.org 0x0
reset:
        brai    kernel_entry
.org 0x8
exception:
        brai    exception       # crash
.org 0x10
interrupt:
        brai    interrupt_handler
.org 0x18
break:
        brai    break           # crash
.org 0x20
hw_exception:
        brai    hw_exception    # crash
.org REGION_CHANGE
region_change_vector:
        bra     r3

.org 0x50
.align 2
interrupt_handler:
        # Save registers not saved by call_function
        addik   r1, r1, -switch_area_size
        swi     r3, r1, r3_offset
        swi     r4, r1, r4_offset
        swi     r5, r1, r5_offset
        swi     r6, r1, r6_offset
        swi     r7, r1, r7_offset
        swi     r8, r1, r8_offset
        swi     r9, r1, r9_offset
        swi     r10,r1, r10_offset
        swi     r11,r1, r11_offset
        swi     r12,r1, r12_offset
        swi     r14,r1, r14_offset
        swi     r15,r1, r15_b_offset
        swi     r18,r1, r18_offset
        swi     r19,r1, r19_offset

        # Where shall we call next?
        brlid   r15, scheduler
        nop

        or      r19, r3, r0     # TCB pointer copied to non-volatile
        beqid   r3, do_not_task_switch
        or      r5, r0, r0
        lwi     r5, r3, 0       # get task function

        # Notify: a new task will begin
        beqid   r5, task_func_is_zero
        .long   CMD_NTFB

        # Call that task
        brlid   r15, call_function
        lwi     r6, r19, 8       # get block info word

        # Task exit
        swi     r0, r19, 4      # clear running flag
        lwi     r5, r19, 0      # get task function

        # Preempt immediately on return from ISR in case we have 
        # another task waiting to run
measurement_patch_point:
        cput    r0, rfsl1

        # Notify: task finished
        .long   CMD_TFIN

do_not_task_switch:
        # Notify: returning from interrupt
        .long   CMD_RTFI
        lwi     r3, r1, r3_offset
        lwi     r4, r1, r4_offset
        lwi     r5, r1, r5_offset
        lwi     r6, r1, r6_offset
        lwi     r7, r1, r7_offset
        lwi     r8, r1, r8_offset
        lwi     r9, r1, r9_offset
        lwi     r10,r1, r10_offset
        lwi     r11,r1, r11_offset
        lwi     r12,r1, r12_offset
        lwi     r14,r1, r14_offset
        lwi     r15,r1, r15_b_offset
        lwi     r18,r1, r18_offset
        lwi     r19,r1, r19_offset

        # Return from interrupt
        rtid    r14, 0
        addi    r1, r1, switch_area_size


kernel_idle:
        # Notify: idle task start
        .long   CMD_IDLE

        # Clock start - preempt as soon as interrupts are enabled
        # The clock has to be enabled in SYSTEM mode for the sake of the 
        # simulator which assumes that stopped clock == SYSTEM.
        cput    r0, rfsl1

        # USER mode: enable interrupts, breaks
        msrset  r5, MSR_IE | MSR_BE

        # Idle loop
0:      .long   CMD_HALT # HALT (nop in hardware)
        bri     0b

_open:  rtsd    r15, 8
        nop

.text 
    .align 2

init_asm_start:
        # SYSTEM mode (with caches, or not, as appropriate):
        mts     rmsr, r3

        # Kernel entry point, set up globals
        addik   r1, r0, SYSTEM_STACK
        addik   r17, r0, __end_spm

        # Early message
        addik   r5, r0, msg_early
        bralid  r15, print
        nop

        #bralid  r15, mtest
        #nop

        # Fill SPM and block stack with valid blocks
        addik   r5, r0, init_c_start
        addik   r6, r0, DEFAULT_CALL_BIW
        bralid  r15, call_function
        nop

        # Should never reach here because init_c_start does not return
        braid   exit
        addi    r5, r0, 99

notify_task_name:
        # r5 = task name
        # r6 = function ptr
        rtsd    r15, 8
        .long   CMD_NAME

_exit:
exit:   # SYSTEM mode: disable interrupts
        msrclr  r3, MSR_IE 
        addi    r5, r0, exit_text
        bralid  r15, print
        .long   CMD_EXIT

0:      bri     0b          # exit loop

.text
.align BLOCK_SIZE_SHIFT
cs_measure_ep:
        # Maximum of 8 words in this subroutine
        get     r5, rfsl1
        swi     r5, r0, cs_measure_sample
        rtsd    r15, 8
        nop

.align BLOCK_SIZE_SHIFT
init_c_start:
        # SYSTEM mode: interrupts off
        msrclr  r3, MSR_IE
        addi    r5, r0, tick_text
        bralid  r15, print
        nop

        # r17 space is scarce so reset this (this function never returns)
        #addik   r17, r0, __end_spm

        # Initialise C runtime
        addi	r6, r0, __bss_start             /* clear BSS */
        addi	r7, r0, __bss_end		
        rsub	r18, r6, r7		
        blei	r18, Lendbss
Lloopbss:	
        swi	    r0, r6, 0
        addi	r6, r6, 4
        rsub	r18, r6, r7
        bgti	r18, Lloopbss
Lendbss:

        addik   r5, r0, msg_main
        bralid  r15, print
        nop

        # Call main()
        addi    r5, r0, 0
        addi    r6, r0, 0
        addi    r15, r0, exit - 8
        braid	main
        addi    r7, r0, 0

.data
.align 2
cs_measure_sample:
        .long 0
exit_text:
        .ascii  "\r\nexit\xfe\r\n\0"
tick_text:
        .ascii ":"
        .byte 0
msg_early:
        .ascii  "Carousel OS"
        .byte 0
msg_main:
        .ascii  " main()\n"
        .byte 0



